/*******************************
* KM-40 web written by Katsumi *
*   This script is released    *
*     under the LGPL v2.1.     *
*******************************/
/*
	Public methods:
	km40cpu.reset();
	km40cpu.setSpeed(clk);
	km40cpu.exec(msec);
	
	Required outside methods:
	Methods written in z80functions.js.
	memory.read(addr);
	memory.write(addr,data);
	km40cpu.events(); // optional
*/

/*
CPU specifications:

This is a big-endinan CPU.

Address: 12 bits
Data: 4 bits

Total 16384 bits (2048 bytes)

Resisters:
	A:	4 bits (Accumulator)
	B:	4 bits (General register)
	C:	4 bits (General register)
	H:	4 bits (Index high)
	L:	4 bits (Index low)
	F:	2 bits (flags: C and Z)
	S:  4 bits (segment register)
	PC:	12 bits (program counter)
	SP: 6 bits (stack pointer)

Memory access by index (HL):
     <------12------->
     <-4-> <-4-> <-4->
	+-----+-----+-----+
	|  S  |  H  |  L  |
    +-----+-----+-----+

Code access by JP command:
     <------12------->
     <-4-> <---7--->
	+-----+---------+-+
	|  S  | literal |0|
    +-----+---------+-+

Mnemonic

0x00-0x06:	LD A,x (x=-, B, C, (HL), H, L, or S)
0x00     :	NOP
0x07     :  POP A
0x08-0x0e:	LD B,x (x=A, -, C, (HL), H, L, or S)
0x09     :	??? (HALT?)
0x0f     :  POP B
0x10-0x16:	LD C,x (x=A, B, -, (HL), H, L, or S)
0x12     :	RET
0x17     :  POP C
0x18-0x1e:	LD (HL),x (x=A, B, C, -, H, L, or S)
0x1b     :	PUSH PC
0x1f     :  POP F
0x20-0x26:	LD H,x (x=A, B, C, (HL), -, L, or S)
0x24     :	if NC
0x27     :  POP H
0x28-0x2e:	LD L,x (x=A, B, C, (HL), H, -, or S)
0x2d     :	if C
0x2f     :  POP L
0x30-0x36:	LD S,x (x=A, B, C, (HL), H, L, or -)
0x36     :	if NZ
0x37     :  POP S
0x38-0x3e:  PUSH x (x=A, B, C, (HL), H, L, or S)
0x3f     :  if Z
0x40-0x47:	inc x (x=A, B, C, (HL), H, L, S, SP)
0x48-0x4f:	dec x (x=A, B, C, (HL), H, L, S, SP)
0x50-0x5f:  LD S,xx
0x60-0x63:	ADD A,x (x=A, B, C, or (HL))
0x64-0x67:	SUB A,x (x=A, B, C, or (HL))
0x68-0x6b:	ADC A,x (x=A, B, C, or (HL))
0x6c-0x6f:	SBC A,x (x=A, B, C, or (HL))
0x70-0x73:	AND A,x (x=A, B, C, or (HL))
0x74-0x77:	CP  A,x (x=A, B, C, or (HL))
0x78-0x7b:	OR  A,x (x=A, B, C, or (HL))
0x7c-0x7f:	XOR A,x (x=A, B, C, or (HL))
0x80-0xff:	Jump with 7 bits operand

Notes:

1) Skip flag may be set after "if" command.
If it is set, the next operation is skipped (recognized as a "NOP").

2) "LD S,xx" followed by "JP"/"LD x,S" can be recognized as 2 byte commands.

3) "SL" (shift left) is the same as "ADD A".
"RL" (rotate left) is the same as "ADC A".
"SR" (shift right) is the same as "ADD A", "ADC A", "ADC A", "ADC A".
"RR" (rotate right) is the same as "ADC A", "ADC A", "ADC A", "ADC A".

4) "PUSH PC" command pushes 12 bit address as 3 words stacks.
When "RET" command is executed, skip flag is set.
This makes CALL operation simpler.
Generally, "JP" command follows "CALL" command and this "JP" will be skipped after "RET".

5) "JP C,xxxx" can be written as follows: "LD S,xx", "IF C", "JP xx"
"CALL xxxx" can be written as follows: "LD S,xx", "PUSH PC", "JP xx"
"CALL C,xxxx" cannot be used, so far.
*/

km40cpu=new Object();
km40cpu.mnemonic=[
	"NOP","LD A,B","LD A,C","LD A,(HL)","LD A,H","LD A,L","LD A,S","POP A",
	"LD B,A","???","LD B,C","LD B,(HL)","LD B,H","LD B,L","LD B,S","POP B",
	"LD C,A","LD C,B","RET","LD C,(HL)","LD C,H","LD C,L","LD C,S","POP C",
	"LD (HL),A","LD (HL),B","LD (HL),C","PUSH PC","LD (HL),H","LD (HL),L","LD (HL),S","POP F",
	"LD H,A","LD H,B","LD H,C","LD H,(HL)","IF NC","LD H,L","LD H,S","POP H",
	"LD L,A","LD L,B","LD L,C","LD L,(HL)","LD L,H","IF C","LD L,S","POP L",
	"LD S,A","LD S,B","LD S,C","LD S,(HL)","LD S,H","LD S,L","IF NZ","POP S",
	"PUSH A","PUSH B","PUSH C","PUSH F","PUSH H","PUSH L","PUSH S","IF Z",
	"INC A","INC B","INC C","INC (HL)","INC H","INC L","INC S","INC SP",
	"DEC A","DEC B","DEC C","DEC (HL)","DEC H","DEC L","DEC S","DEC SP",
	"LD S,00","LD S,01","LD S,02","LD S,03","LD S,04","LD S,05","LD S,06","LD S,07",
	"LD S,08","LD S,09","LD S,0A","LD S,0B","LD S,0C","LD S,0D","LD S,0E","LD S,0F",
	"ADD A","ADD B","ADD C","ADD (HL)","SUB A","SUB B","SUB C","SUB (HL)",
	"ADC A","ADC B","ADC C","ADC (HL)","SBC A","SBC B","SBC C","SBC (HL)",
	"AND A","AND B","AND C","AND (HL)","CP A","CP B","CP C","CP (HL)",
	"OR A","OR B","OR C","OR (HL)","XOR A","XOR B","XOR C","XOR (HL)",
	"JP 00","JP 02","JP 04","JP 06","JP 08","JP 0A","JP 0C","JP 0E",
	"JP 10","JP 12","JP 14","JP 16","JP 18","JP 1A","JP 1C","JP 1E",
	"JP 20","JP 22","JP 24","JP 26","JP 28","JP 2A","JP 2C","JP 2E",
	"JP 30","JP 32","JP 34","JP 36","JP 38","JP 3A","JP 3C","JP 3E",
	"JP 40","JP 42","JP 44","JP 46","JP 48","JP 4A","JP 4C","JP 4E",
	"JP 50","JP 52","JP 54","JP 56","JP 58","JP 5A","JP 5C","JP 5E",
	"JP 60","JP 62","JP 64","JP 66","JP 68","JP 6A","JP 6C","JP 6E",
	"JP 70","JP 72","JP 74","JP 76","JP 78","JP 7A","JP 7C","JP 7E",
	"JP 80","JP 82","JP 84","JP 86","JP 88","JP 8A","JP 8C","JP 8E",
	"JP 90","JP 92","JP 94","JP 96","JP 98","JP 9A","JP 9C","JP 9E",
	"JP A0","JP A2","JP A4","JP A6","JP A8","JP AA","JP AC","JP AE",
	"JP B0","JP B2","JP B4","JP B6","JP B8","JP BA","JP BC","JP BE",
	"JP C0","JP C2","JP C4","JP C6","JP C8","JP CA","JP CC","JP CE",
	"JP D0","JP D2","JP D4","JP D6","JP D8","JP DA","JP DC","JP DE",
	"JP E0","JP E2","JP E4","JP E6","JP E8","JP EA","JP EC","JP EE",
	"JP F0","JP F2","JP F4","JP F6","JP F8","JP FA","JP FC","JP FE"
];
km40cpu.assemble=new Object();
for (i=0;i<256;i++) {
	km40cpu.assemble[km40cpu.mnemonic[i]]=i;
}
km40cpu.read=function(code){
	switch (code) {
		case 'A':
			return this.regA;
		case 'B':
			return this.regB;
		case 'C':
			return this.regC;
		case '(HL)':
			return memory.read((this.regS<<8)|(this.regH<<4)|this.regL);
		case 'H':
			return this.regH;
		case 'L':
			return this.regL;
		case 'S':
			return this.regS;
		case 'F':
			return this.regF;
		default:
			if (code.match(/^0[0-9A-F]$/)) return parseInt(code,16);
			throw "Syntax error:"+code;
	}
};
km40cpu.write=function(code,data){
	data&=0x0f;
	switch (code) {
		case 'A':
			this.regA=data;
			return;
		case 'B':
			this.regB=data;
			return;
		case 'C':
			this.regC=data;
			return;
		case '(HL)':
			memory.write((this.regS<<8)|(this.regH<<4)|this.regL,data);
			return;
		case 'H':
			this.regH=data;
			return;
		case 'L':
			this.regL=data;
			return;
		case 'S':
			this.regS=data;
			return;
		case 'F':
			this.regF=data;
			return;
		default:
			if (!re) throw "Syntax error:"+code;
	}
};
km40cpu.NOP=function(){ };
km40cpu.LD=function(code){
	var re=code.match(/^([^,]+),([^,]+)$/);
	if (!re) throw "Syntax error:"+code;
	this.write(re[1],this.read(re[2]));
};
km40cpu.setFlags=function(result){
	if (result&0x10) {
		this.regF=0x01; // Set carry
	} else {
		this.regF=0x00; // Clear carry
	}
	if ((result&0x0f)==0x00) {
		this.regF|=0x02; // Set zero
	}
};
km40cpu.INC=function(code){
	if (code=='SP') {
		this.regSP++;
		this.regSP&=0x0fff;
		this.regSP|=0x0fc0;
		return;
	}
	var result=this.read(code)+1;
	this.write(code,result&0x0f);
	this.setFlags(result);
};
km40cpu.DEC=function(code){
	if (code=='SP') {
		this.regSP+=63;
		this.regSP&=0x0fff;
		this.regSP|=0x0fc0;
		return;
	}
	var result=(this.read(code)+0x1f)&0x1f;
	this.write(code,result&0x0f);
	this.setFlags(result);
};
km40cpu.push=function(data){
	this.DEC('SP');
	memory.write(this.regSP,data);
};
km40cpu.pop=function(){
	var data=memory.read(this.regSP);
	this.INC('SP');
	return data;
};
km40cpu.PUSH=function(code){
	if (code=="PC") {
		this.push(this.regPC&0x0f);
		this.push((this.regPC>>4)&0x0f);
		this.push((this.regPC>>8)&0x0f);
	} else {
		this.push(this.read(code));
	}
};
km40cpu.POP=function(code){
	this.write(code,this.pop());
};
km40cpu.RET=function(){
	this.regPC=this.pop()<<8;
	this.regPC|=this.pop()<<4;
	this.regPC|=this.pop();
	this.flagSkip=true;
};
km40cpu.IF=function(code){
	switch (code) {
		case 'C':
			if (!(this.regF&0x01)) this.flagSkip=true;
			return;
		case 'NC':
			if ((this.regF&0x01)) this.flagSkip=true;
			return;
		case 'Z':
			if (!(this.regF&0x02)) this.flagSkip=true;
			return;
		case 'NZ':
			if ((this.regF&0x02)) this.flagSkip=true;
			return;
		default:
			if (!re) throw "Syntax error:"+code;
	}
};
km40cpu.ADD=function(code){
	var result=this.regA+this.read(code);
	this.regA=result&0x0f;
	this.setFlags(result);
};
km40cpu.SUB=function(code){
	var result=(32+this.regA-this.read(code))&0x1f;
	this.regA=result&0x0f;
	this.setFlags(result);
};
km40cpu.ADC=function(code){
	var result=this.regA+this.read(code)+(this.regF&0x01);
	this.regA=result&0x0f;
	this.setFlags(result);
};
km40cpu.SBC=function(code){
	var result=(32+this.regA-this.read(code)-(this.regF&0x01))&0x1f;
	this.regA=result&0x0f;
	this.setFlags(result);
};
km40cpu.AND=function(code){
	var result=this.regA&this.read(code);
	this.regA=result&0x0f;
	this.setFlags(result);
};
km40cpu.CP=function(code){
	var result=(32+this.regA-this.read(code))&0x1f;
	this.setFlags(result);
};
km40cpu.OR=function(code){
	var result=this.regA|this.read(code);
	this.regA=result&0x0f;
	this.setFlags(result);
};
km40cpu.XOR=function(code){
	var result=this.regA^this.read(code);
	this.regA=result&0x0f;
	this.setFlags(result);
};
km40cpu.JP=function(code){
	this.regPC=(this.regS<<8)|parseInt(code,16);
};
km40cpu.execOne=function(code){
	if (this.flagSkip) {
		this.flagSkip=false;
		return;
	}
	var mnemonic=this.mnemonic[code];
	var op1=mnemonic.match(/^[\S]+/);
	var op2=mnemonic.match(/[\S]+$/);
	try {
		this[op1[0]](op2[0]);
	} catch(e) {
		var err="Error when executing ";
		err+=("0"+code.toString(16)).substr(-2)+"("+mnemonic+")\n";
		err+=e+"\n";
		err+="Line "+e.lineNumber+" in "+e.fileName.replace(/^.*?([^\/\\\\]+)$/,"$1");
		alert(err);
	}
};
km40cpu.reset=function(){
	this.regA=0;
	this.regB=0;
	this.regC=0;
	this.regH=0;
	this.regL=0;
	this.regS=0;
	this.regF=0;
	this.regPC=0;
	this.regSP=0x0fc0;
	this.flagSkip=false;
};
km40cpu.events=function(){
	// This will be called every msec.
};
km40cpu.speed=1000; // Default: 1 MHz
km40cpu.setSpeed=function(clk){
	this.speed=Math.ceil(clk/1000);
};
km40cpu.counter=1;
km40cpu.exec=function(msec){
	var i,code;
	msec=msec*this.speed;
	for (i=0;i<msec;i++) {
		if (this.regPC==this.breakPoint) {
			this.breakPoint=-1;
			this.step=1;
			return;
		}
		code=memory.read(this.regPC++)<<4;
		code|=memory.read(this.regPC++);
		if (0xfff<this.regPC) this.regPC=0;
		this.execOne(code);
		if ((--this.counter)<1) {
			this.counter=this.speed;
			this.events();
		}
		if (this.step) return;
	}
};
